# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Bit fields data store type."""

from typing import Dict, List, Optional, Set, Tuple, Type, TypeVar, Union

from .._getbytes import getbytes
from ..data import Data
from ..dump import Record
from ..enum import EnumX
from ..exceptions import InsufficientMemoryError
from .bitfield import BitField, bitfield
from .meta import BitFieldsMeta, ByteOrderHint

BitField.__module__ = "plum.bitfields"
bitfield.__module__ = "plum.bitfields"
BitFieldsMeta.__module__ = "plum.bitfields"

CLS = TypeVar("CLS", bound="BitFields")


class BitFields(
    Data, metaclass=BitFieldsMeta, nbytes=4, byteorder="little", default=0, ignore=0
):

    """Bit fields data store type."""

    # filled in by metaclass
    __byteorder__: ByteOrderHint = "little"
    __compare_mask__: int = 0xFFFFFFFF
    __fields__: Dict = dict()
    __default__: int = 0
    __fields_with_defaults__: Set = set()
    __ignore__: int = 0
    __max__: int = 0xFFFFFFFF
    __nbytes__: int = 4

    def __init__(self):
        self.__value__ = self.__default__

    @classmethod
    def from_int(cls: Type[CLS], value: int) -> CLS:
        """Create instance populated from integer value."""
        instance = object.__new__(cls)

        if not 0 <= value <= cls.__max__:
            raise ValueError(
                f"{cls.__name__}.from_int() requires 0 <= value <= {cls.__max__}"
            )

        instance.__value__ = int(value)

        return instance

    @classmethod
    def __pack__(
        cls,
        value: Union["BitFields", int],
        pieces: List[bytes],
        dump: Optional[Record] = None,
    ) -> None:
        if isinstance(value, dict):
            value = cls(**value)

        if dump is None:
            pieces.append(
                int(value).to_bytes(cls.__nbytes__, cls.__byteorder__, signed=False)
            )
        else:
            dump.value = str(value)  # in case something goes wrong

            int_value = int(value)
            chunk = int_value.to_bytes(cls.__nbytes__, cls.__byteorder__, signed=False)

            pieces.append(chunk)

            dump.value = str(int_value)
            dump.memory = chunk
            cls.__add_bitfields_to_dump__(int_value, dump)

    @classmethod
    def __unpack__(
        cls, buffer: bytes, offset: int, dump: Optional[Record] = None
    ) -> Tuple["BitFields", int]:
        if dump is None:
            end = offset + cls.__nbytes__

            if len(buffer) < end:
                raise InsufficientMemoryError("too few bytes to unpack")

            int_value = int.from_bytes(
                buffer[offset:end], cls.__byteorder__, signed=False
            )

            return cls.from_int(int_value), end

        chunk, offset = getbytes(buffer, offset, dump, cls.__nbytes__)

        int_value = int.from_bytes(chunk, cls.__byteorder__, signed=False)

        bitfields = cls.from_int(int_value)

        dump.value = bitfields.__value__
        cls.__add_bitfields_to_dump__(bitfields, dump)

        return bitfields, offset

    @classmethod
    def __add_bitfields_to_dump__(cls, bitfields, dump, bitoffset=0):
        if not isinstance(bitfields, cls):
            bitfields = cls.from_int(int(bitfields))

        for name, accessor in cls.__fields__.items():
            bitfield_typ = accessor.typ

            if not isinstance(bitfield_typ, EnumX) and issubclass(
                bitfield_typ, BitFields
            ):
                row = dump.add_record(access=name, fmt=bitfield_typ)
                bitfield_typ.__add_bitfields_to_dump__(
                    getattr(cls, name).__get__(bitfields, cls),
                    row,
                    bitoffset + accessor.lsb,
                )

            else:
                dump.add_record(
                    access=name,
                    bits=(bitoffset + accessor.lsb, accessor.size),
                    value=getattr(cls, name).__get__(bitfields, cls),
                    fmt=bitfield_typ,
                )

    @classmethod
    def _normalize_for_compare(cls, value, other):
        if isinstance(other, dict):
            other = cls(**other)

        if isinstance(other, cls):
            other = other.__value__ & cls.__compare_mask__
            value = value & cls.__compare_mask__
        else:
            try:
                other = int(other)
            except ValueError:
                pass

        return value, other

    def __lt__(self, other):
        value, other = self._normalize_for_compare(self.__value__, other)
        return int.__lt__(value, other)

    def __le__(self, other):
        value, other = self._normalize_for_compare(self.__value__, other)
        return int.__le__(value, other)

    def __eq__(self, other):
        value, other = self._normalize_for_compare(self.__value__, other)
        return int.__eq__(value, other)

    def __ne__(self, other):
        value, other = self._normalize_for_compare(self.__value__, other)
        return int.__ne__(value, other)

    def __gt__(self, other):
        value, other = self._normalize_for_compare(self.__value__, other)
        return int.__gt__(value, other)

    def __ge__(self, other):
        value, other = self._normalize_for_compare(self.__value__, other)
        return int.__ge__(value, other)

    def __hash__(self):
        return int.__hash__(self.__value__ & type(self).__compare_mask__)

    def __bool__(self):
        return int.__bool__(self.__value__ & type(self).__compare_mask__)

    def __add__(self, other):
        return int.__add__(self.__value__, other)

    def __sub__(self, other):
        return int.__sub__(self.__value__, other)

    def __mul__(self, other):
        return int.__mul__(self.__value__, other)

    def __truediv__(self, other):
        return int.__truediv__(self.__value__, other)

    def __floordiv__(self, other):
        return int.__floordiv__(self.__value__, other)

    def __mod__(self, other):
        return int.__mod__(self.__value__, other)

    def __divmod__(self, other):
        return int.__divmod__(self.__value__, other)

    def __pow__(self, other, *args):
        return int.__pow__(self.__value__, other, *args)

    def __lshift__(self, other):
        return int.__lshift__(self.__value__, other)

    def __rshift__(self, other):
        return int.__rshift__(self.__value__, other)

    def __and__(self, other):
        return int.__and__(self.__value__, other)

    def __xor__(self, other):
        return int.__xor__(self.__value__, other)

    def __or__(self, other):
        return int.__or__(self.__value__, other)

    def __radd__(self, other):
        return int.__radd__(self.__value__, other)

    def __rsub__(self, other):
        return int.__rsub__(self.__value__, other)

    def __rmul__(self, other):
        return int.__rmul__(self.__value__, other)

    def __rtruediv__(self, other):
        return int.__rtruediv__(self.__value__, other)

    def __rfloordiv__(self, other):
        return int.__rfloordiv__(self.__value__, other)

    def __rmod__(self, other):
        return int.__rmod__(self.__value__, other)

    def __rdivmod__(self, other):
        return int.__rdivmod__(self.__value__, other)

    def __rpow__(self, other, *args):
        return int.__rpow__(self.__value__, other, *args)

    def __rlshift__(self, other):
        return int.__rlshift__(self.__value__, other)

    def __rrshift__(self, other):
        return int.__rrshift__(self.__value__, other)

    def __rand__(self, other):
        return int.__rand__(self.__value__, other)

    def __rxor__(self, other):
        return int.__rxor__(self.__value__, other)

    def __ror__(self, other):
        return int.__ror__(self.__value__, other)

    def __iadd__(self, other):
        return self.__value__ + other

    def __isub__(self, other):
        return self.__value__ - other

    def __imul__(self, other):
        return self.__value__ * other

    def __itruediv__(self, other):
        return self.__value__ / other

    def __ifloordiv__(self, other):
        return self.__value__ // other

    def __imod__(self, other):
        return self.__value__ % other

    def __ilshift__(self, other):
        return self.__value__ << other

    def __irshift__(self, other):
        return self.__value__ >> other

    def __iand__(self, other):
        return self.__value__ & other

    def __ixor__(self, other):
        return self.__value__ ^ other

    def __ior__(self, other):
        return self.__value__ | other

    def __neg__(self):
        return -self.__value__

    def __pos__(self):
        return self.__value__

    def __abs__(self):
        return self.__value__

    def __invert__(self):
        return ~self.__value__

    def __int__(self):
        return self.__value__

    def __float__(self):
        return int.__float__(self.__value__)

    def __index__(self):
        return int.__index__(self.__value__)

    def __round__(self, *args):
        return int.__round__(self.__value__, *args)

    def asdict(self):
        """Return bit field values in dictionary form.

        :returns: bit field names/values
        :rtype: dict

        """
        return {name: getattr(self, name) for name in self.__fields__}

    def __setattr__(self, key, value):
        if key.startswith("__") or key in self.__fields__:
            super().__setattr__(key, value)
        else:
            raise AttributeError(f"{type(self).__name__!r} has no attribute {key!r}")

    def __getitem__(self, index):
        nbits = self.__nbytes__ * 8
        mask = 1
        bits = []
        value = int(self)
        while nbits:
            bits.append(bool(value & mask))
            mask <<= 1
            nbits -= 1
        return bits[index]

    def __setitem__(self, index, value):
        bits = self[:]
        nbits = len(bits)
        bits[index] = value
        if len(bits) != nbits:
            raise ValueError("slice and value not same length")
        i = 0
        mask = 1
        for bit in bits:
            if bit:
                i |= mask
            mask <<= 1
        self.__value__ = i



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test array transform support for structure's dimmed_member."""

from plum.array import ArrayX
from plum.bigendian import uint8
from plum.dump import Dump


class TestUnpack:

    """Test unpack method variants."""

    buffer = b"\x00\x01\x02"
    greedy = ArrayX(name="greedy", fmt=uint8)

    def test_unpack(self):
        array, _offset = self.greedy.__unpack__(self.buffer, 0, dims=(2,))
        assert array == [0, 1]

    def test_unpack_and_dump(self):
        array, _offset = self.greedy.__unpack_and_dump__(
            self.buffer, 0, Dump(), dims=(2,)
        )
        assert array == [0, 1]



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Sample BitFields data store class."""

from enum import IntEnum
from typing import Union

from plum.bitfields import bitfield, BitFields
from plum.enum import EnumX


class Register(IntEnum):

    """Tested class."""

    PC = 0
    SP = 1
    R0 = 2
    R1 = 3


class MyBits(BitFields, nbytes=2, default=0x0):

    """Sample BitFields subclass."""

    f1: int = bitfield(lsb=0, size=8, default=0xAB)
    f2: Register = bitfield(lsb=8, size=4, typ=Register)
    f3: bool = bitfield(lsb=12, size=1, typ=bool)
    f4: Union[int, Register] = bitfield(lsb=13, size=2, typ=EnumX(Register))



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2022 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""IPv4/Ipv6 object to bytes and bytes to IPv4/Ipv6 object transform."""

from ipaddress import IPv4Address
from typing import Any, Callable, Union

from ._typing import ByteOrderHint
from .object import ObjectX

PackType = Callable[[Any], bytes]
UnpackType = Callable[[bytes], Any]


class IPv4AddressX(ObjectX):

    """IPv4Address to bytes and bytes to IPv4Address transform."""

    def __init__(
        self,
        byteorder: ByteOrderHint = "big",
        name: str = "IPv4Address",
    ) -> None:
        nbytes = 4

        def pack(ipv4address: Union[int, str, IPv4Address]) -> bytes:
            """Pack IPv4 address as formatted bytes.

            :raises: ``PackError`` if type error, value error, etc.

            """
            return int(IPv4Address(ipv4address)).to_bytes(nbytes, byteorder)

        def unpack(buffer: bytes) -> IPv4Address:
            """Unpack IPv4Address from formatted bytes.

            :raises: ``UnpackError`` if insufficient bytes, excess bytes, or value error

            """
            return IPv4Address(int.from_bytes(buffer, byteorder))

        super().__init__(pack, unpack, nbytes, name, hint="IPv4Address")

        self.__byteorder__: ByteOrderHint = byteorder

    @property
    def byteorder(self) -> ByteOrderHint:
        """Byte order ("little" or "big")."""
        return self.__byteorder__



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test Structure type inheritance behaviors."""

from baseline import Baseline

from plum.littleendian import uint8, uint16
from plum.structure import Structure, member


class Base(Structure):

    m0 = member(fmt=uint8)
    m1 = member(fmt=uint8)


class TestAddMethod:

    """Inherit members, but just add a method."""

    class Struct(Base):
        @property
        def product(self):
            return self.m0 * self.m1

    def test_init_and_property(self):
        assert self.Struct(m0=2, m1=3).product == 6

    def test_round_trip(self):
        expected_dump = Baseline(
            """
            +--------+--------+-------+-------+--------------------+
            | Offset | Access | Value | Bytes | Format             |
            +--------+--------+-------+-------+--------------------+
            |        |        |       |       | Struct (Structure) |
            | 0      | m0     | 1     | 01    | uint8              |
            | 1      | m1     | 2     | 02    | uint8              |
            +--------+--------+-------+-------+--------------------+
            """
        )
        original = self.Struct(m0=1, m1=2)
        buffer, pack_dump = original.ipack_and_dump()
        copy, unpack_dump = self.Struct.unpack_and_dump(buffer)
        assert str(pack_dump) == expected_dump
        assert str(unpack_dump) == expected_dump
        assert original == copy


class TestAddMembers:

    """Inherit members, but just add a method."""

    class Struct(Base):

        m2 = member(fmt=uint8)
        m3 = member(fmt=uint8)

    def test_round_trip(self):
        expected_dump = Baseline(
            """
            +--------+--------+-------+-------+--------------------+
            | Offset | Access | Value | Bytes | Format             |
            +--------+--------+-------+-------+--------------------+
            |        |        |       |       | Struct (Structure) |
            | 0      | m0     | 1     | 01    | uint8              |
            | 1      | m1     | 2     | 02    | uint8              |
            | 2      | m2     | 3     | 03    | uint8              |
            | 3      | m3     | 4     | 04    | uint8              |
            +--------+--------+-------+-------+--------------------+
            """
        )
        original = self.Struct(m0=1, m1=2, m2=3, m3=4)
        buffer, pack_dump = original.ipack_and_dump()
        copy, unpack_dump = self.Struct.unpack_and_dump(buffer)
        assert str(pack_dump) == expected_dump
        assert str(unpack_dump) == expected_dump
        assert original == copy


class TestOverrideDefaults:

    """Inherit, overriding the format and default values for some members."""

    class Base(Structure):
        base_default_val = member(fmt=uint8, default=0)
        base_default_factory = member(fmt=uint8, default=lambda self: 1)

        to_override_base_default_val = member(fmt=uint8, default=2)
        to_override_base_default_factory = member(fmt=uint8, default=lambda self: 3)

        order_outer_0 = member(fmt=uint8, default=4)
        order_inner = member(fmt=uint8, default=5)
        order_outer_1 = member(fmt=uint8, default=6)

    class Inheriting(Base):
        # Make sure untouched fields (base_default_val, base_default_factory) inherit properly

        # Override format and defaults
        to_override_base_default_val = member(fmt=uint16, default=20)
        to_override_base_default_factory = member(fmt=uint16, default=lambda self: 30)

        # Ensure order is preserved when overriding a single member
        order_inner = member(fmt=uint16, default=lambda self: 50)

        # Add new fields with both types of defaults
        new_0 = member(fmt=uint16, default=70)
        new_1 = member(fmt=uint16, default=lambda self: 80)

    def test_inheriting(self):
        """Test structure with inherited and overriding members."""
        expected_dump = Baseline(
            """
            +--------+----------------------------------+-------+-------+------------------------+
            | Offset | Access                           | Value | Bytes | Format                 |
            +--------+----------------------------------+-------+-------+------------------------+
            |        |                                  |       |       | Inheriting (Structure) |
            |  0     | base_default_val                 | 0     | 00    | uint8                  |
            |  1     | base_default_factory             | 1     | 01    | uint8                  |
            |  2     | to_override_base_default_val     | 20    | 14 00 | uint16                 |
            |  4     | to_override_base_default_factory | 30    | 1e 00 | uint16                 |
            |  6     | order_outer_0                    | 4     | 04    | uint8                  |
            |  7     | order_inner                      | 50    | 32 00 | uint16                 |
            |  9     | order_outer_1                    | 6     | 06    | uint8                  |
            | 10     | new_0                            | 70    | 46 00 | uint16                 |
            | 12     | new_1                            | 80    | 50 00 | uint16                 |
            +--------+----------------------------------+-------+-------+------------------------+
            """
        )
        inheriting = self.Inheriting()
        _, dump = inheriting.ipack_and_dump()
        assert str(dump) == expected_dump

    def test_specify_arg_vals(self):
        """Test structure with specified (non-default) values for some inherited and new members."""
        expected_dump = Baseline(
            """
            +--------+----------------------------------+-------+-------+------------------------+
            | Offset | Access                           | Value | Bytes | Format                 |
            +--------+----------------------------------+-------+-------+------------------------+
            |        |                                  |       |       | Inheriting (Structure) |
            |  0     | base_default_val                 | 255   | ff    | uint8                  |
            |  1     | base_default_factory             | 100   | 64    | uint8                  |
            |  2     | to_override_base_default_val     | 200   | c8 00 | uint16                 |
            |  4     | to_override_base_default_factory | 300   | 2c 01 | uint16                 |
            |  6     | order_outer_0                    | 4     | 04    | uint8                  |
            |  7     | order_inner                      | 500   | f4 01 | uint16                 |
            |  9     | order_outer_1                    | 6     | 06    | uint8                  |
            | 10     | new_0                            | 700   | bc 02 | uint16                 |
            | 12     | new_1                            | 800   | 20 03 | uint16                 |
            +--------+----------------------------------+-------+-------+------------------------+
            """
        )
        inheriting = self.Inheriting(
            base_default_val=255,
            base_default_factory=100,
            to_override_base_default_val=200,
            to_override_base_default_factory=300,
            order_inner=500,
            new_0=700,
            new_1=800,
        )
        _, dump = inheriting.ipack_and_dump()
        assert str(dump) == expected_dump
